/* http://www.jslint.com/lint.html */
/*jslint */
/*global prompt : true,
         console : true,
         setTimeout : true,
         document : true
*/

/**
 * Needed separately for jslint validation.
 */
var mpp;

/**
 * Global namespace for Mobile Payments Platform JS Bridge.
 *
 * <p/><strong>Events:</strong>
 * <ul>
 * <li><code>document#mpp-ready</code> - after successful initialization
 * </ul>
 *
 * <p/><strong>Note:</strong>
 * <ul>
 * <li>Android JSBridge must be initialized before script execution.</li>
 * </ul>
 *
 * @global
 * @namespace mpp
 */
mpp = mpp || (function () {
    "use strict";

    /**
     * Callback functions for currently executing operations.
     *
     * @memberof mpp
     * @member {Object.<string,function>} callbackFunctions
     * @inner
     * @private
     */
    var mpp_callbackFunctions = {
        },

        /**
         * Settings for callback
         *
         * @memberof mpp
         * @member {Object.<string,function>} callbackSettings
         * @property {number} nextCallbackSerialNumber
         * @inner
         * @private
         */
        mpp_callbackSettings = {
            nextCallbackSerialNumber : 0
        },

        /**
         * Debugging property to enable logging.
         *
         * @memberof mpp
         * @member {boolean} isLoggingAllowed
         * @inner
         * @private
         * @since 1
         */
        mpp_isLoggingAllowed = false,

        /**
         * Wrapper debugging console.
         *
         * @memberof mpp
         * @member {Object} console
         * @inner
         * @private
         * @since 1
         */
        mpp_console = (function () {
            var mpp_consoleLog = function (method, args) {
                if (!mpp_isLoggingAllowed) {
                    return;
                }
                try {
                    console[method].apply(console, args);
                } catch (ignore) {
                }
            };

            return {
                debug : function () {
                    mpp_consoleLog("debug", arguments);
                },
                info : function () {
                    mpp_consoleLog("info", arguments);
                },
                warn : function () {
                    mpp_consoleLog("warn", arguments);
                },
                error : function () {
                    mpp_consoleLog("error", arguments);
                }
            };
        }()),

        /**
         * Generates serial callback id
         *
         * @function getCallbackId
         * @memberof mpp
         * @inner
         * @private
         *
         * @returns {unique callback id}
         */
        mpp_getCallbackId = function () {
            var serialNumber = mpp_callbackSettings.nextCallbackSerialNumber;
            mpp_callbackSettings.nextCallbackSerialNumber += 1;
            return serialNumber + "_" + Date.now().toString(36);
        },

        /**
         * Safely serializes exceptions.
         *
         * @function serializeException
         * @memberof mpp
         * @inner
         * @private
         * @param {Object} e - Thrown exception to serialize.
         * @return {String}
         */
        mpp_serializeException = function (e) {
            var key,
                copy;
            try {
                return JSON.stringify(e);
            } catch (ignore) {
            }

            copy = {};
            for (key in e) {
                if (e.hasOwnProperty(key)) {
                    // Be extra safe that only keys which can be serialized are copied
                    try {
                        JSON.stringify({ v : e[key] });
                        copy[key] = e[key];
                    } catch (ignore) {
                    }
                }
            }
            return JSON.stringify(copy);
        },

        /**
         * List available plugins and operations.
         *
         * @memberof mpp
         * @function listPlugins
         * @since 1
         *
         * @return {Object} See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/PluginInfoList.java">PluginInfoList.java</a>.
         */
        mpp_listPlugins = function () {
            return JSON.parse(prompt("MPP_listPlugins"));
        },

        /**
         * Safely calls JSON's stringify on an object, throwing an error if the object does not exist.
         * This prevents the error case of a null object being stringified into the actual string 'null'
         *
         * @memberof mpp
         * @function safeStringify
         * @inner
         * @private
         *
         * @param {Object} object - to be serialized into JSON string.
         * @returns {Object} JSON string serialization of input object.
         */
        mpp_safeStringify = function (object) {
            if (!object) {
                throw new Error("mpp-init.js: mpp_safeStringify: Empty object passed!");
            }
            return JSON.stringify(object);
        },

        /**
         * Invokes operation with subsequent callback if invocation is successful.
         *
         * @memberof mpp
         * @function invokePlugin
         * @since 1
         *
         * @param {Function} callbackFunction - Executed if invoke is successful and after execution completion. Function must take two arguments <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationExecuteResponse.java">OperationExecuteResponse.java</a> and operationResponse. See plugin response API for specifics on operationResponse.
         * @param {Object} operation - See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/Operation.java">Operation.java</a>.
         * @param {Object} operationRequest - See plugin request API for specifics on operationRequest.
         * @returns {Object} See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationInvokeResponse.java">OperationInvokeResponse.java</a>.
         */
        mpp_invokePlugin = function (callbackFunction, operation, operationRequest) {
            if ((typeof callbackFunction) !== "function") {
                throw new Error("Callback function must be a function. {type: [" + (typeof callbackFunction) + "]}");
            }

            var callbackId = mpp_getCallbackId(),
                isPluginResponseExpected = true,
                payloadJson,
                invokeResponseJson,
                invokeResponse;

            mpp_callbackFunctions[callbackId] = {
                func : callbackFunction,
                startMillis : Date.now()
            };

            try {
                payloadJson =
                    mpp_safeStringify({
                        callbackId : callbackId,
                        operation : operation,
                        // For now, operationRequest MUST be in JSON for plugins to work:
                        operationRequestJson : mpp_safeStringify(operationRequest)
                    });
                invokeResponseJson = prompt("MPP_invokePlugin_" + payloadJson);
                invokeResponse = JSON.parse(invokeResponseJson);
                if (invokeResponse.error) {
                    isPluginResponseExpected = false;
                }
                return invokeResponse;
            } catch (e) {
                isPluginResponseExpected = false;
                throw e;
            } finally {
                if (!isPluginResponseExpected) {
                    delete mpp_callbackFunctions[callbackId];
                }
            }
        },

        /**
         * Creates function to execute callback.
         *
         * @memberof mpp
         * @function createCallback
         * @inner
         * @private
         *
         * @param {String} callbackId - Identifier to callback function.
         * @param {JsonString} executeResponseJson - See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationExecuteResponse.java">OperationExecuteResponse.java</a>.
         * @param {JsonString} operationResponseJson - See plugin resposne API for specifics on operationResponse.
         * @returns {Function}
         */
        mpp_createCallback = function (callbackId, executeResponseJson, operationResponseJson) {
            var callbackFunction = mpp_callbackFunctions[callbackId],
                executeResponse = JSON.parse(executeResponseJson),
                operationResponse;
            delete mpp_callbackFunctions[callbackId];

            if (operationResponseJson) {
                operationResponse = JSON.parse(operationResponseJson);
            } else {
                if (!executeResponse.error) {
                    mpp_console.warn("Operation response was blank and execute wasn't an error for callback: [%s]", callbackId);
                }
                operationResponse = null;
            }

            return function () {
                var durationMillis = Date.now() - callbackFunction.startMillis,
                    payloadJson;
                try {
                    callbackFunction.func(executeResponse, operationResponse);
                    payloadJson = mpp_safeStringify(
                            {
                                callbackId : callbackId,
                                isSuccess : true,
                                exceptionJson : null // success;
                            });
                    prompt("MPP_callbackResponse_" + payloadJson);
                    mpp_console.debug("Callback succeeded: id: [%s], durationMillis: [%s]", callbackId, durationMillis);
                } catch (e) {
                    mpp_console.error("Failed callback: [" + callbackId + "], durationMillis: [" + durationMillis + "]", e);
                    payloadJson = mpp_safeStringify(
                            {
                                callbackId : callbackId,
                                isSuccess : false,
                                exceptionJson : mpp_serializeException(e) // failure;
                            });
                    prompt("MPP_callbackResponse_" + payloadJson);
                }
            };
        },

        /**
         * Initiates the callback function after operation execution has completed.
         *
         * @memberof mpp
         * @function firePluginCallback
         * @private
         *
         * @param {String} callbackId - Identifier to callback function.
         * @param {JsonString} executeResponseJson - See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationExecuteResponse.java">OperationExecuteResponse.java</a>.
         * @param {JsonString} operationResponseJson - See plugin response API for specifics on operationResponse.
         * @returns {void}
         */
        mpp_firePluginCallback = function (callbackId, executeResponseJson, operationResponseJson) {
            var payloadJson;
            try {
                // Prevent concurrent callbacks
                setTimeout(mpp_createCallback(callbackId, executeResponseJson, operationResponseJson), 0);
            } catch (e) {
                mpp_console.error("Failed callback: [" + callbackId + "]", e);
                payloadJson = mpp_safeStringify(
                        {
                            callbackId : callbackId,
                            isSuccess : false,
                            exceptionJson : mpp_serializeException(e) // failure
                        });
                prompt("MPP_callbackResponse_" + payloadJson);
            }
        },

        /**
         * Payment Provider Eligibility invoke plugin helper method.
         *
         * @memberof mpp
         * @function getPaymentProviderEligibility
         * @since 1
         *
         * @param {Function} callbackFunction - Executed if invoke is successful and after execution completion. Function must take two arguments <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationExecuteResponse.java">OperationExecuteResponse.java</a> and operationResponse. See operationResponse POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/capabilities/v1/GetPaymentProviderEligibilityResponse.java">GetPaymentProviderEligibilityResponse.java</a>.
         * @param {Object} request - (Optional) See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/capabilities/v1/GetPaymentProviderEligibilityRequest.java">GetPaymentProviderEligibilityRequest.java</a>.
         * @returns {Object} See POJO model: <a href="https://code.amazon.com/packages/MPPJavascriptBridgeModel/blobs/mainline/--/src/com/amazon/adapt/mpp/jsbridge/model/OperationInvokeResponse.java">OperationInvokeResponse.java</a>.
         */
        mpp_getPaymentProviderEligibility = function (callbackFunction, request) {
            return mpp_invokePlugin(
                callbackFunction,
                {
                    pluginName : "CapabilitiesPlugin",
                    pluginVersion : 1,
                    pluginOperationName : "GET_PAYMENT_PROVIDER_ELIGIBILITY"
                },
                request || {}
            );
        },

        /**
         * Notify system mpp is ready.
         *
         * @memberof mpp
         * @function ready
         * @private
         * @inner
         *
         * @event document#mpp-ready
         *
         * @returns {void}
         */
        mpp_ready = function () {
            var event, // The custom event that will be created
                eventName = "mpp-ready",
                mppTypeOf = typeof mpp;

            if (mppTypeOf === "undefined") {
                mpp_console.warn("Delaying mpp-ready for mpp to be defined... mppTypeOf: [%s]", mppTypeOf);
                setTimeout(mpp_ready, 0);
                return;
            }

            mpp_console.debug("Firing mpp-ready...");
            // we need to check which event mechanism is on the webview.
            if (document.createEvent) {
                // create event object
                event = document.createEvent("Event");
                event.eventName = eventName;
                event.initEvent(eventName, true, true);

                // fire event
                document.dispatchEvent(event);
            } else {
                // create event object
                event = document.createEventObject();
                event.eventName = eventName;
                event.eventType = eventName;

                //fire event
                document.fireEvent(event.eventType, event);
            }
        },

        /**
         * Glue code the widget calls to invoke a plugin.
         *
         * @memberof mpp
         * @function onWidgetInvokePlugin
         * @inner
         * @private
         *
         * @param {Object} params - Passed by widget. Contains error function, success function, adapt operation to call, and the operation data to use.
         */
        mpp_onWidgetInvokePlugin = function(params) {
            if (typeof params !== 'object') {
                throw new Error("mpp_onWidgetInvokePlugin must be passed params of {expected type: [Object]} {actual type: [" + (typeof params) + "]}");
            }
            //Callback function after invokePlugin has been executed
            var invokePluginCallback = function(executeResponse, operationResponse) {
                    //Asynchronous response from invokePlugin
                    if (executeResponse.error) {
                        params.error();
                    } else {
                        params.success(operationResponse);
                    }
                },
                invokePlugin = function() {
                    var invokeResponse = mpp_invokePlugin(invokePluginCallback, params.adaptOperation, params.operationData);
                    //Synchronous response from invokePlugin
                    if (invokeResponse.error) {
                        params.error();
                    }
                };

            if (typeof window.mpp === 'undefined') {
                document.addEventListener("mpp-ready", invokePlugin);
            } else {
                //Execute invokePlugin immediately because mpp is on the page
                //Using setTimeout method to invokePlugin asynchronously
                setTimeout(invokePlugin, 0);
            }
        },

        /**
         * Clients must register a widget if they intend to allow the widget to invoke a plugin. Widgets that
         * do not require plugin invocation do not need to register a widget.
         *
         * @memberof mpp
         * @function registerWidgetToInvokePlugin
         *
         * @param {Object} widget - Widget that requires a listener for onInvokeExternalOperation.
         */
        mpp_registerWidgetToInvokePlugin = function(widget) {
            if (typeof widget === 'object') {
                widget.oninvokeExternalOperation = mpp_onWidgetInvokePlugin;
            } else {
                throw new Error("oninvokeExternalOperation was not initialized because widget is not defined as an object.");
            }
        },

        /**
         * @memberof mpp
         * @function init
         * @private
         * @inner
         * @version 1
         *
         * @returns {Object} exposed fields
         */
        mpp_init = function () {
            var exposedObject = {
                    proxyVersion : 1
                };

            try {
                /* Exposing for client usage */
                exposedObject.getPaymentProviderEligibility = mpp_getPaymentProviderEligibility;
                exposedObject.invokePlugin = mpp_invokePlugin;
                exposedObject.listPlugins = mpp_listPlugins;
                exposedObject.registerWidgetToInvokePlugin = mpp_registerWidgetToInvokePlugin;

                /* Exposing for bridge callback validation */
                exposedObject.privateFirePluginCallback = mpp_firePluginCallback;

                /* Visible for testing only */
                if (prompt("MPP_isDebuggable") === "true") {
                    mpp_isLoggingAllowed = true;
                    exposedObject.callbackFunctions = mpp_callbackFunctions;
                    exposedObject.callbackSettings = mpp_callbackSettings;
                }

                mpp_console.debug("Initialized mpp");
                // fire an mpp-ready event.
                setTimeout(mpp_ready, 0);
            } catch (e) {
                mpp_console.error("Failed to initialize mpp", e);
            }

            return exposedObject;
        };


    return mpp_init();
}());
